import java.util.*;


// speed of animation
float tick=0.2;

TreeSet<n_ellipse> scene_objects;
n_rect scene_window=new n_rect(new PVector(0,0),800,600);


class n_rect{
  //left bottom corner
  private PVector pos;
  private float w;
  private float h;
  
  public float getWidth(){
    return w;
  }
  
  public float getHeight(){
    return h;
  }
  
  public void setWH(float w,float h){
    this.w=w;
    this.h=h;
  }
  
  public n_rect(PVector pos,float w,float h){
    this.pos=pos;
    this.w=w;
    this.h=h;
  }
  
  public boolean contains(PVector center, float radius){
    if(pos.x<(center.x+radius) && pos.x+w>(center.x-radius) ){
      if(pos.y<(center.y+radius) && pos.y+h>(center.y-radius) ){
        return true;
      }
      return false;
    }
    return false;
  }
  public PVector getNearestPlaneNormal(PVector center){
    float[] a=new float[]{abs(center.x-pos.x),abs(center.x-pos.x-w),abs(center.y-pos.y),abs(center.y-pos.y-h)};
    PVector[] vec= new PVector[]{new PVector(0,1),new PVector(0,-1),new PVector(1,0),new PVector(-1,0)};
    float min=10000000000000000f;
    int min_index=-1;
    for(int i=0;i<4;i++){
      if(min>a[i]){
        min=a[i];
        min_index=i;
      }
    }
    return vec[min_index];
  }
}
PVector Refract(PVector incidentVec, PVector normal, float eta)
{
  PVector out=new PVector();
  float N_dot_I = PVector.dot(normal, incidentVec);
  float k = 1.f - eta * eta * (1.f - N_dot_I * N_dot_I);
  if (k < 0.f)
    out.set(0.f, 0.f, 0.f);
  else
    out =  PVector.sub(incidentVec.mult(eta), normal.mult(eta * N_dot_I + sqrt(k)));
  return out;
}

class n_ellipse implements Comparable{
  private PVector tmp_dir;
  private PVector pos;
  private PVector dir;
  private float w;
  private float h;
  private float move_speed;
  private float w_grow_speed;
  private float h_grow_speed;
  protected final int id;
  private int lifetime=0;
  private final int period; 
  private color rgb;
  
  public PVector getPos(){return pos;}
  public float getRadius(){return max(w,h);}
  
  public n_ellipse(PVector pos, float w, float h, PVector dir,float move_speed,float w_grow_speed,float h_grow_speed,int period,color rgb, int id){
    tmp_dir=new PVector();
    this.pos=pos;
    this.dir=dir.normalize();
    this.w=w;
    this.h=h;
    this.move_speed=move_speed;
    this.w_grow_speed=w_grow_speed;
    this.h_grow_speed=h_grow_speed;
    this.id=id;
    this.period=period;
    this.rgb=rgb;
  }
  // updates variables
  public void update(PVector pos, float w, float h, PVector dir, float move_speed, float w_grow_speed,float h_grow_speed){
    this.pos=pos;
    this.dir=dir.normalize();
    this.w=w;
    this.h=h;
    this.move_speed=move_speed;
    this.w_grow_speed=h_grow_speed;
  }
  
  // updates transform of ellipse
  // sinc function, could be called after draw or with periodic time lapse
  public void tick_update(){
    scene_window.setWH(width,height);
    
    tmp_dir.set(dir);
    tmp_dir.mult(tick*move_speed);
    pos.add(tmp_dir);

    w=w+w_grow_speed*tick;
    h=h+h_grow_speed*tick;
    lifetime++;
    if(lifetime==period){
      lifetime=0;
      this.w=-this.w;
      this.h=-this.h;
    }
  }
  
  public void draw(){
    fill(rgb,100);
    ellipse(pos.x,pos.y,w,h);
  }
  
  public int compareTo(Object other){
    if(((n_ellipse)other).id==id)
      return 0;
    else if(((n_ellipse)other).id<id)
      return -1;
    return 1;
  }  
  //changes dirrection of ellipse
  public void collides(PVector normal){
      float incidenceVectorX = dir.x;
      float incidenceVectorY = dir.y;
      float dot = incidenceVectorX*normal.x + incidenceVectorY*normal.y;
      // calculate reflection vector
      dir.x = 2*normal.x*dot - incidenceVectorX;
      dir.y = 2*normal.y*dot - incidenceVectorY;
      dir.normalize();
  }
}

void setup() {
  size(800, 600,P2D);
  colorMode(RGB, 255);
  scene_objects=new TreeSet();
  for(int i=0;i<1000;i++)
    scene_objects.add(new n_ellipse(new PVector(random(0+10,scene_window.getWidth()-10),random(0+10,scene_window.getHeight()-10)),  //pos
                                    random(50,100),                                                  //h
                                    random(50,100),                                                  //w
                                    new PVector(random(-10,10),random(-10,10)),                      //move_direction
                                    random(-2,2),                                                    //move_speed
                                    random(-1,1),                                                    //grow_speed_w
                                    random(-0.5,0.5),                                                //grow_speed_h
                                    (int)random(50,100),                                             //period of object in ticks
                                    color(random(50,250),random(50,100),random(150,250)),
                                    i));                                                             //id
  smooth();
  noStroke();
  background(0);
}

void keyPressed() {
  if (keyCode == UP) {
    if(tick<3f)
      tick+=0.1;
  } else if (keyCode == DOWN) {
    if(tick>0)
      tick-=0.1;
  }
  System.out.println(tick);
}

void draw() {
  background(0);
  for(n_ellipse i : scene_objects){
    i.draw();        
    i.tick_update(); // updates possition of ellipses
    
    // collision handling
    if(!scene_window.contains(i.getPos(),i.getRadius())){
      i.collides(scene_window.getNearestPlaneNormal(i.getPos()));
    }
  }
}